package fri.util.javascript;

import java.io.IOException;
import java.io.OutputStream;
import java.io.PrintStream;
import java.net.URL;
import java.util.ArrayList;
import java.util.Date;
import java.util.HashMap;
import java.util.LinkedHashMap;
import java.util.List;
import java.util.Map;

/**
 * Expects output as produced by bomTest.js. Turns that output
 * into a test statistic when <code>getResultAsHtml()</code> is
 * called after execution of JS.
 * An instance of this can be set to <code>ConsoleMock</code>
 * as write-stream.
 * 
 * @author Fritz Ritzberger Oct 7, 2014
 */
public class BomTestPrintStream extends PrintStream
{
	private static final String SUCCESS = "SUCCESS: ";	// hardcoded in bomTest.js
	private static final String ERROR = "ERROR: ";	// hardcoded in bomTest.js
	
	private static class BrowserResult
	{
		public final Map<String,String> results = new LinkedHashMap<String,String>();
		public final Map<String,Boolean> success = new HashMap<String,Boolean>();
	}
	
	public final Map<String,BrowserResult> browserResults = new LinkedHashMap<String,BrowserResult>();
	private BrowserResult currentBrowserResult;
	
	public BomTestPrintStream(OutputStream out) {
		super(out);
	}

	public void println(Object o) {
		println(""+o);
	}
	
	public void println(String s) {
		if (s.startsWith(SUCCESS))
			tokenizeLogLine(s, SUCCESS, true);
		else if (s.startsWith(ERROR))
			tokenizeLogLine(s, ERROR, false);
		
		super.println(s);
	}
	
	private void tokenizeLogLine(String line, String startToken, boolean wasSuccess) {
		if (currentBrowserResult == null)
			throw new IllegalStateException("tokenizeLogLine expects a browser version to have been set!");
		
		final int comma = line.indexOf(',');
		final String name = line.substring(startToken.length(), comma);
		final String value = wasSuccess ? line.substring(comma + 2 + "value = ".length()) : line.substring(comma + 2);
		
		if (currentBrowserResult.results.containsKey(name) == false)	{
			currentBrowserResult.results.put(name, value);
			currentBrowserResult.success.put(name, Boolean.valueOf(wasSuccess));
		}
	}

	public void setBrowserVersion(String browser) {
		final BrowserResult browserResult = browserResults.get(browser);
		if (browserResult != null)
			throw new IllegalArgumentException("Browser already regsitered: "+browser);
		
		browserResults.put(browser, currentBrowserResult = new BrowserResult());
	}

	/** @return HTML text containing the test result. */
	public String getResultAsHtml(URL scriptUrl) throws IOException	{
		final StringBuilder sb = new StringBuilder();
		
		buildHead(sb);
		
		final Map<String,Boolean> allBrowserSuccessProperties = findAllBrowserSuccessProperties();
		final Map<String,Boolean> allBrowserErrorProperties = findAllBrowserErrorProperties();
		
		// overview table
		buildOverviewTable(sb, allBrowserSuccessProperties, allBrowserErrorProperties);
		
		// detail tables
		buildDetailTables(sb, allBrowserSuccessProperties, allBrowserErrorProperties);
		
		buildScriptAppendix(scriptUrl, sb);
		
		sb.append("<hr/><p style='text-align: center;'>&copy; Fritz Ritzberger, 2014</p>\n");
		sb.append("</body>\n</html>");
		
		return sb.toString();
	}

	private void buildHead(final StringBuilder sb) {
		sb.append("<!DOCTYPE html>\n<html>\n");
		sb.append("<head>\n<title>HtmlUnit JavaScript Support</title>\n");
		sb.append("<meta name='viewport' content='initial-scale=1'/>\n");
		sb.append("<style type='text/css'>\n");
		sb.append("table { border-collapse: collapse; border: 1px solid black; }\n");
		sb.append("td { border: 1px solid black; font-size: smaller; }\n");
		sb.append(".error { color: orange; }\n");
		sb.append(".success { color: blue; }\n");
		sb.append(".all { color: green; }\n");
		sb.append(".none { color: red; }\n");
		sb.append("</style>\n");
		sb.append("</head>\n");
		
		sb.append("<body style='font-family: sans-serif;'>\n<h1>HtmlUnit 2.16 JavaScript Support</h1>\n");
		sb.append("<p>Generated at "+new Date()+"</p>\n");
		sb.append("<hr/>\n");
		sb.append("<p>This is a mix of "+count()+" <a href='https://developer.mozilla.org/en-US/docs/Web/API/window'>mozilla</a> and <a href='http://www.w3schools.com/jsref/obj_window.asp'>w3schools</a> properties and functions, tested against <a href='http://htmlunit.sourceforge.net/'>HtmlUnit</a> 2.16.<p/>\n");
		
		sb.append("<table style='display: inline-block;'><tbody>\n");
		sb.append("<tr><td>Chrome</td><td>Google Chrome</td></tr>\n");
		sb.append("<tr><td>FF</td><td>Mozilla Firefox</td></tr>\n");
		sb.append("<tr><td>IE</td><td>Microsoft Internet Explorer</td></tr>\n");
		sb.append("</tbody></table>\n");
		
		sb.append("<table style='display: inline-block; float: right;'><tbody>");
		sb.append("<tr><td class='all'>GREEN: Property/function exists in all browsers, HtmlUnit test succeeded in all.</td></tr>\n");
		sb.append("<tr><td class='success'>BLUE: Property/function exists in this browsers, HtmlUnit test succeeded.</td></tr>\n");
		sb.append("<tr><td class='error'>ORANGE: Property/function does not exists in this browser, HtmlUnit test failed.</td></tr>\n");
		sb.append("<tr><td class='none'>RED: Property/function does not exists in any browser, HtmlUnit test failed in all.</td></tr>\n");
		sb.append("</tbody></table>\n");
		
		sb.append("<h2>Overview</h2>\n");
	}

	private int count() {
		return browserResults.entrySet().iterator().next().getValue().results.size();
	}

	private void buildOverviewTable(final StringBuilder sb, Map<String,Boolean> allBrowserSuccessProperties, Map<String,Boolean> allBrowserErrorProperties) {
		sb.append("<table>\n");
		sb.append("<thead>\n");
		for (String browser : browserResults.keySet())
			sb.append("<th>"+browser+"</th>");
		sb.append("</thead>\n");
		sb.append("<tbody>\n");
		final List<List<String>> columns = new ArrayList<List<String>>();
		for (Map.Entry<String,BrowserResult> e : browserResults.entrySet())	{
			final String browser = e.getKey();
			final BrowserResult browserResult = e.getValue();
			final List<String> currentColumn = new ArrayList<String>();
			columns.add(currentColumn);
			for (Map.Entry<String,String> e2 : browserResult.results.entrySet())	{
				final String property = e2.getKey();
				final String cssClass = getCssClass(allBrowserSuccessProperties, allBrowserErrorProperties, browserResult, property);
				currentColumn.add("<td title='"+browser+"'><a class='"+cssClass+"' href='"+mozillaUrl(property)+"'>"+property+"</a></td>");
			}
		}
		for (int row = 0; row < columns.get(0).size(); row++)	{	// loop over all properties = rows
			sb.append("<tr>");
			for (int column = 0; column < columns.size(); column++)	// loop over all browsers = columns
				sb.append(columns.get(column).get(row));
			sb.append("</tr>\n");
		}
		sb.append("</tbody>\n");
		sb.append("</table>\n");
	}

	private void buildDetailTables(final StringBuilder sb, Map<String,Boolean> allBrowserSuccessProperties, Map<String,Boolean> allBrowserErrorProperties) {
		sb.append("<h2>Details</h2>\n");
		
		for (Map.Entry<String,BrowserResult> e : browserResults.entrySet())	{
			final String browser = e.getKey();
			final BrowserResult browserResult = e.getValue();
			
			sb.append("<h3>"+browser+"</h3>\n");
			sb.append("<table>\n");
			sb.append("<thead><tr><th>Property/Function</th><th>Value/Return</th></tr></thead>\n");
			sb.append("<tbody>\n");
			for (Map.Entry<String,String> e2 : browserResult.results.entrySet())	{
				final String property = e2.getKey();
				final String value = e2.getValue();
				final String cssClass = getCssClass(allBrowserSuccessProperties, allBrowserErrorProperties, browserResult, property);
				
				sb.append("<tr><td><a class='"+cssClass+"' href='"+mozillaUrl(property)+"'>"+property+"</td><td>"+escape(value)+"</td></tr>\n");
			}
			sb.append("</tbody>\n");
			sb.append("</table>\n");
		}
	}

	private String escape(String value) {
		if (value.indexOf('<') < 0)
			return value;
		return escape(value, true);
	}

	private String escape(String value, boolean andNewlines) {
		final StringBuilder sb = new StringBuilder();
		for (int i = 0; i < value.length(); i++)	{
			final char c = value.charAt(i);
			if (c == '<')
				sb.append("&lt;");
			else if (c == '>')
				sb.append("&gt;");
			else if (c == '&')
				sb.append("&amp;");
			else if (andNewlines && c == '\n')
				sb.append("\\n");
			else if (andNewlines && c == '\r')
				sb.append("\\r");
			else
				sb.append(c);
		}
		return sb.toString();
	}

	private void buildScriptAppendix(URL scriptUrl, final StringBuilder sb) throws IOException {
		if (scriptUrl != null)	{
			sb.append("<h2>JavaScript Test Code</h2>\n");
			sb.append("<pre>\n");
			sb.append(escape(ReaderUtil.toString(scriptUrl), false));
			sb.append("</pre>\n");
		}
	}

	private String getCssClass(Map<String, Boolean> allBrowserSuccessProperties, Map<String, Boolean> allBrowserErrorProperties, BrowserResult browserResult, String property) {
		final boolean success = browserResult.success.get(property).booleanValue();
		final boolean allBrowserSuccess = allBrowserSuccessProperties.get(property).booleanValue();
		final boolean allBrowserError = allBrowserErrorProperties.get(property).booleanValue();
		assert allBrowserSuccess != allBrowserError : "Can not be successful and erroneous in all browsers at the same time: "+ property;
		return allBrowserSuccess ? "all" : allBrowserError ? "none" : success ? "success" : "error";
	}
	
	private Map<String,Boolean> findAllBrowserSuccessProperties() {
		return findAllBrowserSuccessOrErrorProperties(true);
	}

	private Map<String,Boolean> findAllBrowserErrorProperties() {
		return findAllBrowserSuccessOrErrorProperties(false);
	}

	private Map<String,Boolean> findAllBrowserSuccessOrErrorProperties(boolean testForSuccess) {
		final Map<String,Boolean> allBrowserProperties = new HashMap<String,Boolean>();
		for (Map.Entry<String,BrowserResult> e : browserResults.entrySet())	{
			final BrowserResult browserResult = e.getValue();
			
			for (Map.Entry<String,String> e2 : browserResult.results.entrySet())	{
				final String property = e2.getKey();
				final boolean success = browserResult.success.get(property).booleanValue();
				final Boolean currentValue = allBrowserProperties.get(property);
				
				final boolean value = (currentValue == null)
						? success == testForSuccess
						: (success == testForSuccess && currentValue.booleanValue() == true);
					
				allBrowserProperties.put(
						property,
						Boolean.valueOf(value));
			}
		}
		return allBrowserProperties;
	}

	private String mozillaUrl(String property)	{
		return "https://developer.mozilla.org/en-US/docs/Web/API/"+mozilla(property);
	}
	
	private String mozilla(String property)	{
		final int i = property.indexOf('(');
		if (i <= 0)	{	// is a property
			if (property.endsWith("EventListener"))
				return property.replace(property.substring(0, property.indexOf('.')), "EventTarget");	// Mozilla special URL;
			
			if (property.startsWith("location.") && ! property.endsWith(".assign") && ! property.endsWith(".reload") && ! property.endsWith(".replace"))
				return property.replace("location.", "URLUtils.");	// Mozilla special URL
			
			if (property.startsWith("attribute."))
				return property.replace("attribute.", "Attr.");	// Mozilla special URL
			
			if (property.startsWith("ajax."))
				return "XMLHttpRequest";	// Mozilla special URL
			
			return property;
		}
		
		// is a function
		return mozilla(property.substring(0, i));
	}

}
