package fri.util.javascript;

import java.io.BufferedReader;
import java.io.FileNotFoundException;
import java.io.FileReader;
import java.io.InputStreamReader;
import java.io.Reader;
import java.io.StringReader;
import java.net.MalformedURLException;
import java.net.URL;

import javax.script.Bindings;
import javax.script.ScriptContext;
import javax.script.ScriptEngine;
import javax.script.ScriptEngineManager;

/**
 * Encapsulates the JavaScript engine (provided by the Java runtime).
 * 
 * @author Fritz Ritzberger Sep 14, 2014
 */
public class JavaScriptRunner
{
	public JavaScriptRunner() {
		System.err.println("==========================================");
		System.err.println("Application is: "+getClass().getSimpleName());
		System.err.println("Running in: "+System.getProperty("user.dir"));
	}
	
	/**
	 * Executes given JavaScript, given either by URL, file path, or as script.
	 * Allocates a new engine instance for this call.
	 */
	public final void execute(String jsUrlOrFileOrScript) throws Exception {
		execute(new String [] { jsUrlOrFileOrScript });
	}
	
	/**
	 * Executes given JavaScripts, given either by URLs, file paths, or as scripts.
	 * Allocates a new engine instance on each call.
	 */
	public final void execute(String [] jsUrlsOrFilesOrScripts) throws Exception {
		final Object engine = createScriptEngine();
		
		for (String jsUrlOrFileOrScript : jsUrlsOrFilesOrScripts)	{
			Reader reader;
			try	{
				final URL url = new URL(jsUrlOrFileOrScript);
				reader = new InputStreamReader(url.openStream());
			}
			catch (MalformedURLException e1)	{
				try	{
					reader = new FileReader(jsUrlOrFileOrScript);
				}
				catch (FileNotFoundException e2)	{
					try	{
						reader = new InputStreamReader(getClass().getResourceAsStream(jsUrlOrFileOrScript));
					}
					catch (Exception e3)	{
						reader = new StringReader(jsUrlOrFileOrScript);
					}
				}
			}
			
			printExecution(jsUrlOrFileOrScript);
			execute(engine, new BufferedReader(reader), jsUrlOrFileOrScript);
		}
		
		cleanUp();
	}

	/**
	 * Override to use other JS engines.
	 * @return this default implementation returns the javax.script default engine.
	 */
	protected Object createScriptEngine() {
		final ScriptEngineManager engineManager = new ScriptEngineManager();
		final ScriptEngine engine = engineManager.getEngineByMimeType("text/javascript");
		createGlobalBindings(engine);
		printEngine(engine);
		return engine;
	}
	
	/**
	 * Override to use other JS engines.
	 * This default implementation casts the given engine to
	 * javax.script.ScriptEngine and launches it with given reader.
	 */
	protected void execute(Object scriptEngine, Reader reader, String origin) throws Exception	{
		ScriptEngine javaxEngine = (ScriptEngine) scriptEngine;
		javaxEngine.eval(reader);
	}

	/**
	 * Does nothing. Override this to create bindings for a BOM environment.
	 */
	protected void addGlobalBindings(Bindings bindings) {
	}

	/**
	 * Does nothing. Override this to clean up after JavaScript execution.
	 */
	protected void cleanUp() {
	}

	private void createGlobalBindings(ScriptEngine engine) {
		final Bindings bindings = engine.createBindings();
		addGlobalBindings(bindings);
		engine.setBindings(bindings, ScriptContext.GLOBAL_SCOPE);
	}


	private void printEngine(ScriptEngine engine) {
		System.err.println("JavaScript Engine: "+engine);
	}

	private void printExecution(String jsUrlOrFileOrScript) {
		System.err.println("Interpreting JavaScript: "+jsUrlOrFileOrScript);
		System.err.println("==========================================");
	}

}
