package fri.util.javascript;

import java.io.IOException;
import java.util.ArrayList;
import java.util.List;

import javax.script.Bindings;

import com.gargoylesoftware.htmlunit.AlertHandler;
import com.gargoylesoftware.htmlunit.BrowserVersion;
import com.gargoylesoftware.htmlunit.ConfirmHandler;
import com.gargoylesoftware.htmlunit.FailingHttpStatusCodeException;
import com.gargoylesoftware.htmlunit.Page;
import com.gargoylesoftware.htmlunit.PromptHandler;
import com.gargoylesoftware.htmlunit.WebClient;
import com.gargoylesoftware.htmlunit.html.HtmlPage;
import com.gargoylesoftware.htmlunit.javascript.host.Window;

import fri.util.javascript.bom.ConsoleMock;

/**
 * JavaScript files to read and interpret must be given on commandline in order,
 * either as URL or as absolute file path, e.g.:
 * <pre>
 * java RunJavaScript
 * 		file:///media/windows/fri/eclipse-workspace8/javascript/test.html
 * 		file:///media/windows/fri/eclipse-workspace8/javascript/bomTest.js
 * </pre>
 * Necessary HtmlUnit JARs:
 * <pre>
	commons-codec-1.9.jar
	commons-collections-3.2.1.jar
	commons-io-2.4.jar
	commons-lang3-3.3.2.jar
	commons-logging-1.1.3.jar
	cssparser-0.9.14.jar
	htmlunit-2.15.jar
	htmlunit-core-js-2.15.jar
	httpclient-4.3.3.jar
	httpcore-4.3.2.jar
	httpmime-4.3.3.jar
	nekohtml-1.9.21.jar
	sac-1.3.jar
	serializer-2.7.1.jar
	xalan-2.7.1.jar
	xercesImpl-2.11.0.jar
	xml-apis-1.4.01.jar
 * </pre>
 * @author Fritz Ritzberger Sep 24, 2014
 */
public class JavaScriptRunnerWithHtmlUnit extends JavaScriptRunner
{
	public static void main(String [] args) throws Exception {
		new ArgumentRunner(args).run(
			new ArgumentRunner.JavaScriptRunnerFactory() {
				public JavaScriptRunner newJavaScriptRunner(String webPage, BrowserVersion browserVersion)	{
					return new JavaScriptRunnerWithHtmlUnit(webPage, browserVersion);
				}
			}
		);
	}
	
	/**
	 * Combines commandline arguments with HtmlUnit browser versions
	 * and provides a factory for different JavaScriptRunner implementations.
	 * @author Fritz Ritzberger Oct 6, 2014
	 */
	protected static class ArgumentRunner
	{
		public interface JavaScriptRunnerFactory
		{
			JavaScriptRunner newJavaScriptRunner(String webPage, BrowserVersion browserVersion);
		}
		
		private static final BrowserVersion [] browserVersions = new BrowserVersion []	{
			BrowserVersion.getDefault(),
			BrowserVersion.INTERNET_EXPLORER_11,
			BrowserVersion.CHROME,
			BrowserVersion.FIREFOX_24,
			BrowserVersion.FIREFOX_31,
		};
		
		private final List<String> webPages;
		private final String [] javaScripts;
		private boolean once = false;
		
		/**
		 * Analyzes the commandline arguments and prepares for launching them.
		 * In commandline there can be HTML page URLs and JavaScripts.
		 * HTML Pages MUST Be given as URL, scripts can be given also as files or plain text.
		 * @param args the commandline arguments given by the launcher of this application.
		 */
		public ArgumentRunner(String [] args) {
			final List<String> webPages = new ArrayList<String>();
			final List<String> javaScripts = new ArrayList<String>();
			
			for (String arg : args)	{
				final String argument = arg.toLowerCase();
				if (argument.equals("once"))
					once = true;
				else if (argument.endsWith(".js"))
					javaScripts.add(arg);
				else if (argument.endsWith(".html") || argument.endsWith(".htm") || argument.startsWith("http") || argument.startsWith("file:"))
					webPages.add(arg);
				else
					throw new IllegalArgumentException("Unknown extension: "+arg);
			}
			
			if (javaScripts.size() <= 0)
				throw new IllegalStateException("Having no JavaScripts to execute!");
			
			if (webPages.size() <= 0)
				webPages.add("http://www.this-page-intentionally-left-blank.org/");
			
			this.webPages = webPages;
			this.javaScripts = javaScripts.toArray(new String[javaScripts.size()]);
		}
		
		public void run(JavaScriptRunnerFactory factory) throws Exception {
			final BrowserVersion [] browsers = once ? new BrowserVersion [] { browserVersions[0] } : browserVersions;
			for (BrowserVersion browserVersion : browsers)	{
				for (String webPage : webPages)
					factory.newJavaScriptRunner(webPage, browserVersion).execute(javaScripts);
			}
		}
	}
	
	
	protected final WebClient webClient;
	protected final HtmlPage page;
	protected final BrowserVersion browserVersion;
	
	public JavaScriptRunnerWithHtmlUnit(String webPage, BrowserVersion browserVersion) {
		this.browserVersion = (browserVersion != null) ? browserVersion : BrowserVersion.getDefault();
		this.webClient = new WebClient(browserVersion);
		configureHtmlUnit(webClient);
		
		try {
			page = webClient.getPage(webPage);
		}
		catch (FailingHttpStatusCodeException | IOException e) {
			throw new RuntimeException(e);
		}
		System.err.println("HtmlUnit mimicking browser "+webClient.getBrowserVersion());
		System.err.println("Loading web page: "+webPage);
	}
	
	/**
	 * Creates HtmlUnit BOM bindings.
	 */
	protected void addGlobalBindings(Bindings bindings) {
		final Window window = getWindow();
		bindings.put("window", window);
		bindings.put("console", new ConsoleMock());
		//bindings.put("console", window.getConsole());	// Console has no such function "log"
		bindings.put("document", window.getDocument());
		bindings.put("navigator", window.getNavigator());
		bindings.put("location", window.getLocation());
		bindings.put("history", window.getHistory());
		bindings.put("screen", window.getScreen());
	}

	/** @return the HtmlUnit WIndow object for JS context. */
	protected Window getWindow() {
		assert page != null;
		return (Window) page.getEnclosingWindow().getScriptObject();
	}
	
	@Override
	protected void cleanUp() {
		webClient.closeAllWindows();
	}


	private void configureHtmlUnit(WebClient webClient) {
		webClient.setPromptHandler(new PromptHandler() {
			@Override
			public String handlePrompt(Page page, String message) {
				//ConsoleMock.stream.println("PROMPT: "+message);
				return "prompt answer";
			}
		});
		webClient.setAlertHandler(new AlertHandler() {
			@Override
			public void handleAlert(Page page, String message) {
				//ConsoleMock.stream.println("ALERT: "+message);
			}
		});
		webClient.setConfirmHandler(new ConfirmHandler() {
			@Override
			public boolean handleConfirm(Page page, String message) {
				//ConsoleMock.stream.println("CONFIRM: "+message);
				return true;
			}
		});
	}

}
